package parser

import "io"
import "fmt"
import "git.tebibyte.media/sashakoshka/fspl/lexer"
import "git.tebibyte.media/sashakoshka/fspl/errors"

// parser parses tokens from a lexer into syntax entities, which it places into
// a tree.
type parser struct {
	token lexer.Token
	lexer lexer.Lexer
	tree  *Tree
}

// newParser creates a new parser that parses the output of the given lexer.
func newParser (lx lexer.Lexer) (*parser, error) {
	return &parser {
		lexer: lx,
	}, nil
}

// parseInto parses the parser's file into the given syntax tree.
func (this *parser) parseInto (tree *Tree) error {
	this.tree = tree
	err := this.parse()
	if err == io.EOF { err = nil }
	return err
}

// expect checks the current token to see if it matches a list of token kind(s),
// else it returns an error describing what it expected.
func (this *parser) expect (allowed ...lexer.TokenKind) error {
	// fmt.Println("expect", this.token, allowed)
	if !this.token.Is(allowed...) {
		return errors.Errorf (
			this.token.Position, "unexpected %v; expected %s",
			this.token, commaList(allowed...))
	}
	return nil
}

// expectDesc is like expect, but the expected entitie(s) are described
// manually. This can be helpful when a large syntactical entity is expected and
// the first token(s) of it offer insufficient information.
func (this *parser) expectDesc (description string, allowed ...lexer.TokenKind) error {
	// fmt.Println("expectDesc", this.token, description, allowed)
	if !this.token.Is(allowed...) {
		return errors.Errorf (
			this.token.Position, "unexpected %v; expected %s",
			this.token, description)
	}
	return nil
}

// expectNext is like expect, but gets the next token first.
func (this *parser) expectNext (allowed ...lexer.TokenKind) error {
	err := this.next()
	if err != nil { return err }
	// fmt.Println("expectNext", this.token, allowed)
	return this.expect(allowed...)
}

// expectNextDesc is like expectDesc, but gets the next token first.
func (this *parser) expectNextDesc (description string, allowed ...lexer.TokenKind) error {
	err := this.next()
	if err != nil { return err }
	// fmt.Println("expectNextDesc", this.token, description, allowed)
	return this.expectDesc(description, allowed...)
}

// expectValue returns an error if the current token's value does not match the
// allowed values.
func (this *parser) expectValue (kind lexer.TokenKind, allowed ...string) error {
	// fmt.Println("expectValue", this.token, kind, allowed)
	if !((this.token.Is(kind) || kind == 0) && this.token.ValueIs(allowed...)) {
		return errors.Errorf (
			this.token.Position, "unexpected %v; expected %s",
			this.token, commaList(allowed))
	}
	return nil
}

// expectValueDesc is like expectValue, but the expected value(s) are described
// manually.
func (this *parser) expectValueDesc (description string, kind lexer.TokenKind, allowed ...string) error {
	// fmt.Println("expectValueDesc", this.token, description, kind, allowed)
	if !this.token.Is(kind) || !this.token.ValueIs(allowed...) {
		return errors.Errorf (
			this.token.Position, "unexpected %v; expected %s",
			this.token, description)
	}
	return nil
}

func (this *parser) next () error {
	token, err := this.lexer.Next()
	if err != nil { return err }
	this.token = token
	return nil
}

func (this *parser) bug () string {
	return fmt.Sprintln (
		"Bug detected in the compiler!\n" +
		"The parser has taken an unexpected control path.",
		"This could be due to an un-implemented feature.\n" +
		"Please submit a report with this info and stack trace to:",
		"https://git.tebibyte.media/sashakoshka/fspl/issues\n" +
		"The token being parsed was:", this.token)
}

func (this *parser) kind () lexer.TokenKind {
	return this.token.Kind
}

func (this *parser) value () string {
	return this.token.Value
}

func (this *parser) pos () errors.Position {
	return this.token.Position
}

func (this *parser) parse () error {
	err := this.next()
	if err != nil { return err }
	for this.token.Kind != lexer.EOF {
		err = this.parseTopLevel()
		if err != nil { return err }
	}
	return nil
}

func commaList[ELEMENT any] (items ...ELEMENT) string {
	list := ""

	switch {
	case len(items) == 1: list = fmt.Sprint(items[0])
	case len(items) == 2: list = fmt.Sprint(items[0], " or ", items[1])
	default:
		for index, item := range items {
			if index > 0 {
				list += ", "
				if index == len(items) - 1 {
					list += " or "
				}
			}
			list += fmt.Sprintf("%v", item)
		}
	}
	return list
}

func prependCopy[ELEMENT any] (item ELEMENT, array []ELEMENT) []ELEMENT {
	return append([]ELEMENT { item }, array...)
}

func appendCopy[ELEMENT any] (array []ELEMENT, items ...ELEMENT) []ELEMENT {
	newArray := make([]ELEMENT, len(array) + len(items))
	copy(newArray[copy(newArray, array):], items)
	return newArray
}
